#include <stdio.h>
#include <string.h>
#include "md5.h"

void * memcpy_ref (void * dst, const void * src, size_t length) {
int i;

	for (i=0; i < length; i++) {
		((char *) dst)[i] = ((char *) src)[i];
	}
	return dst;
}

void * memcpy_1_asm (void * dst, const void * src, size_t length);
#pragma aux memcpy_1_asm =   \
"       mov eax, edi       " \
"   rep movsb              " \
parm [edi] [esi] [ecx] value [eax];

void * memcpy_1 (void * dst, const void * src, size_t length) {
	return memcpy_1_asm (dst, src, length);
}

void * memcpy_2_asm (void * dst, const void * src, size_t length);
#pragma aux memcpy_2_asm =      \
"     mov edx, edi            " \
"     lea ecx, [edi+edi*2]    " \
"     and ecx, 3              " \
"     sub eax, ecx            " \
"     jle short LEndBytes     " \
" rep movsb                   " \
"     mov ecx, eax            " \
"     and eax, 3              " \
"     shr ecx, 2              " \
" rep movsd                   " \
"     LEndBytes: add ecx, eax " \
" rep movsb                   " \
parm [edi] [esi] [eax] value [edx];

void * memcpy_2 (void * dst, const void * src, size_t length) {
	return memcpy_2_asm (dst, src, length);
}

void * memcpy_3_asm (void * dst, const void * src, size_t length);
#pragma aux memcpy_3_asm =        \
"       mov     edx, edi        " \
"       mov     eax, ecx        " \
"       sub     ecx, edi        " \
"       sub     ecx, eax        " \
"       and     ecx, 7          " \
"       sub     eax, ecx        " \
"       jle     short LEndBytes " \
"       emms                    " \
"   rep movsb                   " \
"       mov     ecx, eax        " \
"       and     eax, 7          " \
"       shr     ecx, 3          " \
"       jz      LEndBytes       " \
"       sub     edi, esi        " \
"l1:	movq	mm0, [esi]      " \
"       movq	[edi+esi], mm0	" \
"       add     esi,8         	" \
"       dec     ecx            	" \
"       jnz     l1             	" \
"       add	edi, esi            " \
"       emms                    " \
"LEndBytes:                     " \
"       add     ecx, eax        " \
"       rep movsb               " \
parm [EDI] [ESI] [ECX]            \
modify [EAX ECX ESI EDI]          \
value [EDX];

void * memcpy_3 (void * dst, const void * src, size_t length) {
	return memcpy_3_asm (dst, src, length);
}

extern void * memcpy_sse (void * dst, const void * src, size_t length);

__int64 tsc (void);
#pragma aux tsc =   \
"      rdtsc      " \
value [edx eax];

void idisable (void);
#pragma aux idisable = "cli";

void ienable (void);
#pragma aux ienable = "sti";

#define SZ (32 * 1024 * 1024)

static char box[SZ];

int test1 (void * mc (void * dst, const void * src, size_t length), const char * s) {
struct MD5Context md5;
unsigned char sig[16];
void * t;
int i;
__int64 t0, t1;

	memset (box, 0, SZ);
	memset (box, 0x5A, 100000);
/*	memset (box, 0x5A, 10000);*/

	idisable ();
	t0 = tsc ();
	t = mc (&box[25600000], box, 6400000);
/*	t = mc (&box[25600001], box, 16000); */
	t1 = tsc ();
	ienable ();

	MD5Init (&md5);
	MD5Update (&md5, box, SZ);
	MD5Final (sig, &md5);

	printf ("%p-", t);
	for (i=0; i < 16; i++) {
		printf ("%02X", sig[i]);
	}
	printf (" %I64d \t", t1 - t0);
	printf (" %s\n", s);


	return 0;
}

#define TH(x) test1 ((x), #x)

int main () {
	TH (memcpy);
	TH (memcpy_ref);
	TH (memcpy_1);
	TH (memcpy_2);
	TH (memcpy_3);
	TH (memcpy_sse);
	return 0;
}

